/*
    This file is part of the KDE libraries
    SPDX-FileCopyrightText: 2000 Stephan Kulow <coolo@kde.org>
    SPDX-FileCopyrightText: 2000-2009 David Faure <faure@kde.org>

    SPDX-License-Identifier: LGPL-2.0-or-later
*/

#ifndef KIO_JOBCLASSES_H
#define KIO_JOBCLASSES_H

// For source compatibility
#include "filecopyjob.h"
#include "filejob.h"
#include "listjob.h"
#include "mimetypejob.h"
#include "multigetjob.h"
#include "simplejob.h"
#include "specialjob.h"
#include "statjob.h"
#include "storedtransferjob.h"
#include "transferjob.h"
#include <kio/mkdirjob.h>

#endif
